# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 tw=100 et ai si
#
# Copyright (C) 2020-2022 Intel Corporation
# SPDX-License-Identifier: BSD-3-Clause
#
# Authors: Artem Bityutskiy <artem.bityutskiy@linux.intel.com>
#          Antti Laakso <antti.laakso@linux.intel.com>
#          Niklas Neronin <niklas.neronin@intel.com>

"""
This module implements CPU properties caching.
"""

from pepclibs import CPUInfo
from pepclibs.helperlibs import ClassHelpers
from pepclibs.helperlibs.Exceptions import ErrorNotFound, ErrorNotSupported

class _PropsCache():
    """
    This class implements properties caching. The cache is indexed by property name and CPU number.
    It takes property scope ("global", "package", etc) into account as well. The cache uses the
    write-through policy.
    """

    def _get_cpus_in_scope(self, cpu, scope):
        """Returns all CPUs sharing the same scope 'scope' as the CPU 'cpu'."""

        if scope == "CPU":
            return (cpu, )

        if scope == "global":
            return self._cpuinfo.get_cpus()

        levels = self._cpuinfo.get_cpu_levels(cpu)
        if scope == "package":
            return self._cpuinfo.package_to_cpus(levels[scope])
        if scope == "die":
            return self._cpuinfo.dies_to_cpus(dies=levels[scope], packages=levels["package"])
        if scope == "node":
            return self._cpuinfo.nodes_to_cpus(nodes=levels[scope], packages=levels["package"])
        if scope == "core":
            return self._cpuinfo.cores_to_cpus(cores=levels[scope], packages=levels["package"])

        raise ErrorNotSupported(f"Scope '{scope}' is not supported")

    def is_cached(self, pname, cpu):
        """
        Checks if '(pname, cpu)' item exists in the cache. Returns 'True' if the item was found and
        'False' otherwise. The argument are as follows.
          * pname - name of the propery.
          * cpu - an integer CPU number.
        """

        if pname in self._cache and cpu in self._cache[pname]:
            return True
        return False

    def get(self, pname, cpu):
        """
        Looks up the '(pname, cpu)' item in the cache. Returns the value if the item was found,
        raises 'ErrorNotFound' otherwise. The argument are as follows.
          * pname - name of the propery.
          * cpu - an integer CPU number.
        """

        try:
            return self._cache[pname][cpu]
        except KeyError:
            raise ErrorNotFound(f"{pname} is not cached for CPU {cpu}") from None

    def remove(self, pname, cpu, scope="CPU"):
        """
        Removes '(pname, cpu)' and all the other items sharing the same scope from the cache.
          * pname - name of the propery.
          * cpu - an integer CPU number.
          * scope - name of scope (e.g. "package", "core").
        """

        cpus = self._get_cpus_in_scope(cpu, scope)

        for cpu in cpus: # pylint: disable=redefined-argument-from-local
            try:
                del self._cache[pname][cpu]
            except KeyError:
                pass

    def add(self, pname, cpu, val, scope="CPU"):
        """
        Add value 'val' for item '(pname, cpu)' to the cache. Add it also for each CPU sharing the
        same scope 'scope'. The argument are as follows.
          * pname - name of the propery.
          * cpu - an integer CPU number.
          * val - value to get cached.
          * scope - name of scope (e.g. "package", "core").
        """

        cpus = self._get_cpus_in_scope(cpu, scope)

        if pname not in self._cache:
            self._cache[pname] = {}
        for cpu in cpus: # pylint: disable=redefined-argument-from-local
            self._cache[pname][cpu] = val

    def __init__(self, cpuinfo=None, pman=None):
        """
        The class constructor. The argument are as follows.
          * pman - the process manager object that defines the host to run the measurements on.
          * cpuinfo - CPU information object generated by 'CPUInfo.CPUInfo()'.
        """

        self._cpuinfo = cpuinfo
        self._close_cpuinfo = cpuinfo is None

        if not self._cpuinfo:
            # 'pman' is only used to initialize 'cpuinfo' in this class.
            self._cpuinfo = CPUInfo.CPUInfo(pman=pman)

        self._cache = {}

    def close(self):
        """Uninitialize the class object."""

        ClassHelpers.close(self, close_attrs=("_cpuinfo",))
